/**
 *  '$RCSfile: ReplMessageHandler.java,v $'
 *    Purpose: A class that handles xml messages passed by the 
 *             replication handler
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Chad Berkley
 *    Release: @release@
 *
 *   '$Author: tao $'
 *     '$Date: 2002/04/27 00:31:52 $'
 * '$Revision: 1.7 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.sql.*;
import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;
import java.util.EmptyStackException;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DeclHandler;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;

/** 
 * A Class implementing callback bethods for the SAX parser to
 * call when processing the XML messages from the replication handler
 */
public class ReplMessageHandler extends DefaultHandler 
{
  private Vector updates = new Vector();
  private Vector indivUpdate = new Vector();
  private Vector indivDelete = new Vector();
  private Vector deletes = new Vector();
  private String server;
  private String dataFile;
  private boolean update = false;
  private boolean delete = false;
  String currentTag = new String();
  
  public ReplMessageHandler()
  {
  }
  
  /**
   * This method starts a new vector for each updatedDocument tag.
   */
  public void startElement(String uri, String localName, String qName, 
                           Attributes attributes) throws SAXException
  {
    currentTag = localName;
    if(localName.equals("updatedDocument"))
    {
      indivUpdate = new Vector();
      update = true;
    }
    else if(localName.equals("deletedDocument"))
    {
      indivDelete = new Vector();
      delete = true;
    }
  }
  
  /**
   * This method write the indivUpdate to updates when it finds the end of
   */
  public void endElement(String uri, String localName, String qName) 
              throws SAXException
  {
    if(localName.equals("updatedDocument"))
    {
      updates.add(new Vector(indivUpdate));
      update = false;
    }
    else if(localName.equals("deletedDocument"))
    {
      deletes.add(new Vector(indivDelete));
      delete = false;
    }
  }
  
  /**
   * Take the data out of the docid and date_updated fields
   */
  public void characters(char[] ch, int start, int length) throws SAXException
  {
    if(currentTag.equals("docid") && update)
    {
      indivUpdate.add(new String(ch, start, length));
    }
    else if(currentTag.equals("docid") && delete)
    {
      indivDelete.add(new String(ch, start, length));
    }
    
    if(currentTag.equals("rev") && update)
    {
      indivUpdate.add(new String(ch, start, length));
      indivUpdate.add(server);
    }
    else if(currentTag.equals("rev") && delete)
    {
      indivDelete.add(new String(ch, start, length));
      indivDelete.add(server);
    }
    
    if(currentTag.equals("date_updated") && update)
    {
      indivUpdate.add(new String(ch, start, length));
      indivUpdate.add(server);
    }
    else if(currentTag.equals("date_updated") && delete)
    {
      indivDelete.add(new String(ch, start, length));
      indivDelete.add(server);
    }
    
    if(currentTag.equals("server"))
    {
      server = new String(ch, start, length);
    }
    
    //Adding data file attribute into indivUpdate vector
    if (currentTag.equals("datafile"))
    {
      dataFile = new String(ch, start, length);
      indivUpdate.add(dataFile);
    }
  }
  
  public Vector getUpdatesVect()
  {
    return updates;
  }
  
  public Vector getDeletesVect()
  {
    return deletes;
  }
}
